from typing import Dict
from cadCAD.configuration import Experiment
from cadCAD.configuration.utils import config_sim
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
import pandas as pd


def select_M_dict(M_dict: Dict[str, object], keys: set) -> Dict[str, object]:
    return {k: v for k, v in M_dict.items()
            if k in keys}


def select_config_M_dict(configs: list, i: int, keys: set) -> Dict[str, object]:
    return select_M_dict(configs[i].sim_config['M'], keys)


def easy_run(state_variables,
             params,
             psubs,
             N_timesteps,
             N_samples,
             use_label=False,
             assign_params=True,
             drop_substeps=True) -> pd.DataFrame:
    """
    Run cadCAD simulations without headaches.
    """

    simulation_parameters = {
        'N': N_samples,
        'T': range(N_timesteps),
        'M': params
    }

    sim_config = config_sim(simulation_parameters)

    from cadCAD import configs
    del configs[:]

    exp = Experiment()
    exp.append_configs(sim_configs=sim_config,
                       initial_state=state_variables,
                       partial_state_update_blocks=psubs)

    from cadCAD import configs
    exec_mode = ExecutionMode()
    exec_context = ExecutionContext(exec_mode.local_mode)
    executor = Executor(exec_context=exec_context, configs=configs)
    (records, tensor_field, _) = executor.execute()

    df = pd.DataFrame(records)

    if drop_substeps == True:
        # Drop all intermediate substeps
        first_ind = (df.substep == 0) & (df.timestep == 0)
        last_ind = df.substep == max(df.substep)
        inds_to_drop = (first_ind | last_ind)
        df = df.loc[inds_to_drop].drop(columns=['substep'])
    else:
        pass

    if assign_params == False:
        pass
    else:
        M_dict = configs[0].sim_config['M']
        params_set = set(M_dict.keys())

        # Logic for getting the assign params criteria
        if type(assign_params) is list:
            selected_params = set(assign_params) & params_set
        elif type(assign_params) is set:
            selected_params = assign_params & params_set
        else:
            selected_params = params_set

        # Attribute parameters to each row
        df = df.assign(**select_config_M_dict(configs, 0, selected_params))
        for i, (_, n_df) in enumerate(df.groupby(['simulation', 'subset', 'run'])):
            df.loc[n_df.index] = n_df.assign(**select_config_M_dict(configs,
                                                                    i,
                                                                    selected_params))

    # Based on Vitor Marthendal (@marthendalnunes) snippet
    if use_label == True:
        psub_map = {order + 1: psub.get('label', '')
                    for (order, psub)
                    in enumerate(psubs)}
        psub_map[0] = 'Initial State'
        df['substep_label'] = df.substep.map(psub_map)

    return df



from .types import Param, SystemParameters, InitialState, Dict, List, ParamSweep
from itertools import product

def sweep_cartesian_product(sweep_params: Dict[str, tuple]) -> Dict[str, tuple]:
    """
    Makes a cartesian product from dictionary values.
    This is useful for plugging inside the sys_params dict, like:
    ```python
    sweep_params = {'a': [0.1, 0.2], 'b': [1, 2]}
    product_sweep 
    sys_params = {**cartesian_product_sweep(sweep_params),
                  'c': [0.1]}
    ```
    Usage:
    >>> sweep_params = {'a': [0.1, 0.2], 'b': [1, 2]}
    >>> cartesian_product_sweep(sweep_params)
    {'a': [0.1, 0.1, 0.2, 0.2], 'b': [1, 2, 1, 2]}
    """
    cartesian_product = product(*sweep_params.values())
    transpose_cartesian_product = zip(*cartesian_product)
    zipped_sweep_params = zip(sweep_params.keys(), transpose_cartesian_product)
    sweep_dict = dict(zipped_sweep_params)
    sweep_dict = {k: list(v) for k, v in sweep_dict.items()}
    return sweep_dict


def prepare_params(params: SystemParameters,
                   cartesian_sweep: bool=False) -> Dict[str, List[object]]:
    simple_params = {k: [v.value]
                     for k, v in params.items()
                     if type(v) is Param}

    sweep_params = {k: v.value
                    for k, v in params.items()
                    if type(v) is ParamSweep}
    if cartesian_sweep is True:
        sweep_params = sweep_cartesian_product(sweep_params)
    else:
        pass
    
    cleaned_params = {**simple_params, **sweep_params}
    return cleaned_params


def prepare_state(state: InitialState) -> Dict[str, object]:
    cleaned_state = {k: v.value
                     for k, v in state.items()}
    return cleaned_state



from typing import Callable, List, Dict, Tuple, Union, NamedTuple

StateVariable = object
Parameter = Tuple[object]

Params = Dict[str, object]
State = Dict[str, object]
Signal = Dict[str, object]
History = List[List[State]]
Substep = int
VariableUpdate = Tuple[str, object]

Policy = Callable[[Params, Substep, History, State], Signal]
StateUpdate = Callable[[Params, Substep, History, State, Signal], VariableUpdate]

StateUpdateBlock = Dict[str, object]
TimestepBlock = List[StateUpdateBlock]

class InitialValue(NamedTuple):
    value: object
    type: type


class Param(NamedTuple):
    value: object
    type: type


class ParamSweep(NamedTuple):
    value: Tuple[object]
    type: type


InitialState = Dict[str, InitialValue]
SystemParameters = Dict[str, Union[Param, ParamSweep]]



from typing import Dict
from cadCAD_tools import easy_run
from cadCAD_tools.types import TimestepBlock, StateVariable, Parameter
from time import time
import pandas as pd


def MEASURE_TIME_SUF(p, s, h, v, p_i): return ('run_time', time())


MEASURING_BLOCK = {
    'label': 'Time Measure',
    'policies': {},
    'variables': {
        'run_time': MEASURE_TIME_SUF
    }
}


def profile_psubs(psubs: TimestepBlock, profile_substeps=True) -> TimestepBlock:
    """
    Updates a TimestepBlock so that a time measuring function is added.
    """
    new_timestep_block = []
    new_timestep_block.append(MEASURING_BLOCK)
    if profile_substeps is True:
        for psub in psubs:
            new_timestep_block.append(psub)
            new_timestep_block.append(MEASURING_BLOCK)
    else:
        pass
    return new_timestep_block


def profile_run(state_variables: Dict[str, StateVariable],
                params: Dict[str, Parameter],
                psubs: TimestepBlock,
                *args,
                profile_substeps=True,
                **kwargs) -> pd.DataFrame:

    if profile_substeps is True:
        kwargs.update(drop_substeps=False)

    new_psubs = profile_psubs(psubs, profile_substeps)
    state_variables.update({'run_time': None})

    return easy_run(state_variables,
                    params,
                    new_psubs,
                    *args,
                    **kwargs)
